@@ -3,11 +3,13 @@ use syn::parse::{Parse, ParseStream};
 use crate::code::Code;
 use crate::help::Help;
 use crate::severity::Severity;
+use crate::url::Url;
 
 pub enum DiagnosticArg {
     Code(Code),
     Severity(Severity),
     Help(Help),
+    Url(Url),
 }
 
 impl Parse for DiagnosticArg {
@@ -19,6 +21,8 @@ impl Parse for DiagnosticArg {
             Ok(DiagnosticArg::Severity(input.parse()?))
         } else if ident == "help" {
             Ok(DiagnosticArg::Help(input.parse()?))
+        } else if ident == "url" {
+            Ok(DiagnosticArg::Url(input.parse()?))
         } else {
             Err(syn::Error::new(
                 ident.span(),
