@@ -1,102 +1,136 @@
-use std::collections::HashMap;
+use std::collections::{HashMap, HashSet};
 
 use proc_macro2::TokenStream;
-use quote::{format_ident, quote, ToTokens};
+use quote::{format_ident, quote};
 use syn::{
+    parenthesized,
     parse::{Parse, ParseStream},
-    punctuated::Punctuated,
     spanned::Spanned,
     Token,
 };
 
-use crate::diagnostic::DiagnosticVariant;
+use crate::fmt;
+use crate::{diagnostic::DiagnosticVariant, fmt::Display};
 
 pub struct Snippets(Vec<Snippet>);
 
 struct Snippet {
-    message: Option<MemberOrString>,
+    message: Option<Display>,
     highlights: Vec<Highlight>,
     source: syn::Member,
     snippet: syn::Member,
 }
 
 struct Highlight {
+    label: Option<Display>,
     highlight: syn::Member,
 }
 
 struct SnippetAttr {
     source: syn::Member,
-    message: Option<MemberOrString>,
+    message: Option<Display>,
 }
 
 struct HighlightAttr {
+    label: Option<Display>,
     snippet: syn::Member,
 }
 
-enum MemberOrString {
-    Member(syn::Member),
-    String(syn::LitStr),
-}
-
-impl ToTokens for MemberOrString {
-    fn to_tokens(&self, tokens: &mut TokenStream) {
-        use MemberOrString::*;
-        match self {
-            Member(member) => member.to_tokens(tokens),
-            String(string) => string.to_tokens(tokens),
-        }
-    }
-}
-
-impl Parse for MemberOrString {
-    fn parse(input: ParseStream) -> syn::Result<Self> {
-        let lookahead = input.lookahead1();
-        if lookahead.peek(syn::Ident) || lookahead.peek(syn::LitInt) {
-            Ok(MemberOrString::Member(input.parse()?))
-        } else if lookahead.peek(syn::LitStr) {
-            Ok(MemberOrString::String(input.parse()?))
-        } else {
-            Err(syn::Error::new(
-                input.span(),
-                "Expected a string or a field reference.",
-            ))
-        }
-    }
-}
-
 impl Parse for SnippetAttr {
     fn parse(input: ParseStream) -> syn::Result<Self> {
-        let punc = Punctuated::<MemberOrString, Token![,]>::parse_terminated(input)?;
-        let span = input.span();
-        let mut iter = punc.into_iter();
-        let source = match iter.next() {
-            Some(MemberOrString::Member(member)) => member,
-            _ => {
+        let source = input.parse::<syn::Member>()?;
+        let message = if input.peek(Token![,]) {
+            input.parse::<Token![,]>()?;
+            let ident = input.parse::<syn::Ident>()?;
+            if ident == "message" {
+                let la = input.lookahead1();
+                if la.peek(syn::token::Paren) {
+                    let content;
+                    parenthesized!(content in input);
+                    if content.peek(syn::LitStr) {
+                        let fmt = content.parse()?;
+                        let args = if content.is_empty() {
+                            TokenStream::new()
+                        } else {
+                            content.parse::<Token![,]>()?;
+                            fmt::parse_token_expr(&content, false)?
+                        };
+                        let display = Display {
+                            fmt,
+                            args,
+                            has_bonus_display: false,
+                        };
+                        Some(display)
+                    } else {
+                        return Err(syn::Error::new(ident.span(), "Invalid argument to message() sub-attribute. The first argument must be a literal string."));
+                    }
+                } else {
+                    input.parse::<Token![=]>()?;
+                    Some(Display {
+                        fmt: input.parse()?,
+                        args: TokenStream::new(),
+                        has_bonus_display: false,
+                    })
+                }
+            } else {
                 return Err(syn::Error::new(
-                    span,
-                    "Source must be an identifier that refers to a Source for this snippet.",
-                ))
+                    ident.span(),
+                    "Invalid sub-attribute. Only `message()` is allowed.",
+                ));
             }
+        } else {
+            None
         };
-        let message = iter.next();
         Ok(SnippetAttr { source, message })
     }
 }
 
 impl Parse for HighlightAttr {
     fn parse(input: ParseStream) -> syn::Result<Self> {
-        let punc = Punctuated::<MemberOrString, Token![,]>::parse_terminated(input)?;
-        let span = input.span();
-        let mut iter = punc.into_iter();
-        let snippet =
-            match iter.next() {
-                Some(MemberOrString::Member(member)) => member,
-                _ => return Err(syn::Error::new(
-                    span,
-                    "must be an identifier that refers to something with a #[snippet] attribute.",
-                )),
-            };
-        Ok(HighlightAttr { snippet })
+        let snippet = input.parse::<syn::Member>()?;
+        let label = if input.peek(Token![,]) {
+            input.parse::<Token![,]>()?;
+            let ident = input.parse::<syn::Ident>()?;
+            if ident == "label" {
+                let la = input.lookahead1();
+                if la.peek(syn::token::Paren) {
+                    let content;
+                    parenthesized!(content in input);
+                    if content.peek(syn::LitStr) {
+                        let fmt = content.parse()?;
+                        let args = if content.is_empty() {
+                            TokenStream::new()
+                        } else {
+                            content.parse::<Token![,]>()?;
+                            fmt::parse_token_expr(&content, false)?
+                        };
+                        let display = Display {
+                            fmt,
+                            args,
+                            has_bonus_display: false,
+                        };
+                        Some(display)
+                    } else {
+                        return Err(syn::Error::new(ident.span(), "Invalid argument to label() sub-attribute. The first argument must be a literal string."));
+                    }
+                } else {
+                    input.parse::<Token![=]>()?;
+                    Some(Display {
+                        fmt: input.parse()?,
+                        args: TokenStream::new(),
+                        has_bonus_display: false,
+                    })
+                }
+            } else {
+                return Err(syn::Error::new(
+                    ident.span(),
+                    "Invalid sub-attribute. Only `label()` is allowed.",
+                ));
+            }
+        } else {
+            None
+        };
+        Ok(HighlightAttr { snippet, label })
     }
 }
 
@@ -143,7 +177,7 @@ impl Snippets {
         for (i, field) in fields.iter().enumerate() {
             for attr in &field.attrs {
                 if attr.path.is_ident("highlight") {
-                    let HighlightAttr { snippet } = attr.parse_args::<HighlightAttr>()?;
+                    let HighlightAttr { snippet, label } = attr.parse_args::<HighlightAttr>()?;
                     if let Some(snippet) = snippets.get_mut(&snippet) {
                         let member = if let Some(ident) = field.ident.clone() {
                             syn::Member::Named(ident)
@@ -153,7 +187,10 @@ impl Snippets {
                                 span: field.span(),
                             })
                         };
-                        snippet.highlights.push(Highlight { highlight: member });
+                        snippet.highlights.push(Highlight {
+                            highlight: member,
+                            label,
+                        });
                     } else {
                         return Err(syn::Error::new(snippet.span(), "Highlight must refer to an existing field with a #[snippet(...)] attribute."));
                     }
@@ -167,29 +204,35 @@ impl Snippets {
         }
     }
 
-    pub(crate) fn gen_struct(&self) -> Option<TokenStream> {
+    pub(crate) fn gen_struct(&self, fields: &syn::Fields) -> Option<TokenStream> {
         let snippets = self.0.iter().map(|snippet| {
             // snippet message
-            let msg = snippet
-                .message
-                .as_ref()
-                .map(|msg| match msg {
-                    MemberOrString::String(str) => {
-                        quote! {
-                            message: std::option::Option::Some(#str),
+            let msg = if let Some(display) = &snippet.message {
+                let members: HashSet<syn::Member> = fields
+                    .iter()
+                    .enumerate()
+                    .map(|(i, field)| {
+                        if let Some(ident) = field.ident.as_ref().cloned() {
+                            syn::Member::Named(ident)
+                        } else {
+                            syn::Member::Unnamed(syn::Index {
+                                index: i as u32,
+                                span: field.span(),
+                            })
                         }
-                    }
-                    MemberOrString::Member(m) => {
-                        quote! {
-                            message: std::option::Option::Some(self.#m.as_ref()),
-                        }
-                    }
-                })
-                .unwrap_or_else(|| {
-                    quote! {
-                        message: std::option::Option::None,
-                    }
-                });
+                    })
+                    .collect();
+                let mut display = display.clone();
+                display.expand_shorthand(&members);
+                let Display { fmt, args, .. } = display;
+                quote! {
+                    message: std::option::Option::Some(format!(#fmt, #args)),
+                }
+            } else {
+                quote! {
+                    message: std::option::Option::None,
+                }
+            };
 
             // Source field
             let src_ident = &snippet.source;
@@ -205,9 +248,20 @@ impl Snippets {
 
             // Highlights
             let highlights = snippet.highlights.iter().map(|highlight| {
-                let Highlight { highlight } = highlight;
-                quote! {
-                    self.#highlight.clone().into()
+                let Highlight { highlight, label } = highlight;
+                if let Some(Display { fmt, args, .. }) = label {
+                    quote! {
+                        (
+                            std::option::Option::Some(
+                                format!(#fmt, #args)
+                            ),
+                            self.#highlight.clone().into()
+                        )
+                    }
+                } else {
+                    quote! {
+                        (std::option::Option::None, self.#highlight.clone().into())
+                    }
                 }
             });
             let highlights = quote! {
@@ -241,33 +295,25 @@ impl Snippets {
             variant.snippets.as_ref().map(|snippets| {
                 let variant_snippets = snippets.0.iter().map(|snippet| {
                     // snippet message
-                    let msg = snippet
-                        .message
-                        .as_ref()
-                        .map(|msg| match msg {
-                            MemberOrString::String(str) => {
-                                quote! {
-                                    message: std::option::Option::Some(#str),
-                                }
-                            }
-                            MemberOrString::Member(m) => {
-                                let m = match m {
-                                    syn::Member::Named(id) => id.clone(),
-                                    syn::Member::Unnamed(syn::Index { index, .. }) => {
-                                        format_ident!("_{}", index)
-                                    }
-                                };
-                                quote! {
-                                    message: std::option::Option::Some(#m.as_ref()),
-                                }
-                            }
-                        })
-                        .unwrap_or_else(|| {
-                            quote! {
-                                message: std::option::Option::None,
-                            }
-                        });
-
+                    let msg = if let Some(display) = &snippet.message {
+                    let members: HashSet<syn::Member> = variant.fields.iter().enumerate().map(|(i, field)| {
+                        if let Some(ident) = field.ident.as_ref().cloned() {
+                            syn::Member::Named(ident)
+                        } else {
+                            syn::Member::Unnamed(syn::Index { index: i as u32, span: field.span() })
+                        }
+                    }).collect();
+                        let mut display = display.clone();
+                        display.expand_shorthand(&members);
+                        let Display { fmt, args, .. } = display;
+                        quote! {
+                            message: std::option::Option::Some(format!(#fmt, #args)),
+                        }
+                    } else {
+                        quote! {
+                            message: std::option::Option::None,
+                        }
+                    };
                     // Source field
                     let src_ident = match &snippet.source {
                         syn::Member::Named(id) => id.clone(),
@@ -293,15 +339,24 @@ impl Snippets {
 
                     // Highlights
                     let highlights = snippet.highlights.iter().map(|highlight| {
-                        let Highlight { highlight } = highlight;
+                        let Highlight { highlight, label } = highlight;
                         let m = match highlight {
                             syn::Member::Named(id) => id.clone(),
                             syn::Member::Unnamed(syn::Index { index, .. }) => {
                                 format_ident!("_{}", index)
                             }
                         };
-                        quote! {
-                            #m.clone().into()
+                        if let Some(Display { fmt, args, ..}) = label {
+                            quote! {
+                                (
+                                    std::option::Option::Some(format!(#fmt, #args)),
+                                    #m.clone().into()
+                                )
+                            }
+                        } else {
+                            quote! {
+                                (std::option::Option::None, #m.clone().into())
+                            }
                         }
                     });
                     let highlights = quote! {
